def mean(points):
    return sum(points) / float(len(points))


def variance(points, m):
    return sum([(x - m) ** 2 for x in points])


def covariance(points, m):
    """Returns the covariance of a list of complex points"""
    covar = 0.0
    for point in points:
        covar += (point.real - m.real) * (point.imag - m.imag)
    return covar


def determination(points, y_mean, b0, b1):
    """Returns the coefficient of determination (R**2) for linear equation and list of complex points"""
    sse = []
    sst = []
    for point in points:
        y_hat = b0 + (b1 * point.real)
        diff = point.imag - y_hat
        sse.append(diff ** 2)
        diff = point.imag - y_mean
        sst.append(diff ** 2)
    if sum(sst) == 0:
        return 1  # Horizontal line
    else:
        return 1 - (sum(sse) / sum(sst))


def coefficients(points):
    """Returns the coefficents of a linear equation and R**2 as a list for a list of complex points"""
    x = [point.real for point in points]
    m = mean(points)
    var = variance(x, m.real)
    if var > 0:
        b1 = covariance(points, m) / var
        b0 = m.imag - b1 * m.real
        r2 = determination(points, m.imag, b0, b1)
        return [b0, b1, r2]
    else:
        return [0, 0, 1]  # Vertical line



from .GraphObj import GraphObj


class UserObject(GraphObj):
    ZONE = 'zone'
    ZONE_PREFIX = 'z'
    ZONE_INDEX = 1

    TYPE = 'type'
    TYPES = ['process', 'entity', 'data store', 'trust zone']
    LABEL = 'label'

    texts = "none"

    def __init__(self, xml, label, sid=None, gid=None, value=None):
        super(UserObject, self).__init__(sid, gid, value)
        self.xml = xml
        self.label = label

    def text_to_mx_value(self):
        value = ""
        last_text = len(self.texts) - 1
        for i, t in enumerate(self.texts):
            style = t.get_mx_style()
            value += "<p style='" + style + "'>" + t.text + "</p>"
            if i != last_text:
                value += "<hr size='1'/>"
        return value
    
    @classmethod
    def get_trust_zone_from_node_label(cls, label):
        if label is not None:
            zone = label.lower().split(cls.ZONE_PREFIX)[cls.ZONE_INDEX]
            if cls._is_valid_trust_zone(zone):
                return zone
        
        return None

    def get_trust_zone(self):
  
        zone = self.xml.get(self.LABEL)
        
        if zone is not None:
            try:
                zone = zone.lower().split(self.ZONE_PREFIX)[self.ZONE_INDEX]
            except IndexError:
                pass
            else:
                if self._is_valid_trust_zone(zone):
                    return zone

        return None

    @classmethod
    def _is_valid_trust_zone(self, zone):
        return(int(zone) <= 9)

    def set_trust_zone(self, zone):
        return self.xml.set(self.ZONE, zone)

    def get_object_type(self):
        type = self.xml.get(self.TYPE)
        if type in self.TYPES:
            return type

    @classmethod 
    def infer_type_from_node(cls, node):

        TRUST_ZONE = 'text;html=1;strokeColor=#82b366;fillColor=#d5e8d4;align=center;verticalAlign=middle;whiteSpace=wrap;overflow=hidden;'
        ELEMENT = 'rounded=0;whiteSpace=wrap;html=1;'
        PROCESS = 'ellipse;whiteSpace=wrap;html=1;aspect=fixed;'
        DATA_STORE = 'shape=partialRectangle;whiteSpace=wrap;html=1;left=0;right=0;fillColor=none;'

        types = {
            'trust zone': TRUST_ZONE,
            'element': ELEMENT,
            'process': PROCESS,
            'data store': DATA_STORE
        
        }
        
        for text in node.texts:
            for key, value in types.items():
                if text.text == value: 
                    return key

        return None


import peewee

db = peewee.SqliteDatabase(":memory:")
    # flows, entities, processes, data store

class BaseModel(peewee.Model):
    class Meta:
        database = db

class Element(BaseModel):
    data = peewee.CharField() # a way to store external system specific entity metadata

class Node(Element):
    identifier = peewee.CharField()
    zone = peewee.IntegerField()
    label = peewee.CharField()
    type = peewee.CharField()
    
class Edge(Element):
    source = peewee.ForeignKeyField(Node, backref='source')
    destination = peewee.ForeignKeyField(Node, backref='destination')
    
    # our graph simplifies all kinds of relationships into flows,
    # and then adds metadata to flows with processes. processes 
    # align with workflows, or user actions, that trigger data flows
    # to occur
    process = peewee.ForeignKeyField(Node, backref='process')

class Threat(BaseModel):
    classification = peewee.CharField()

db.drop_tables([Node, Edge])
db.create_tables([Node, Edge])



import string

class Base():
    feature_bare = string.Template(
"""
Feature: $component
"""        
    )
    feature_base = string.Template(
"""
Feature: $component
    $scenario
"""
    )

class Scenarios():
    stride = {
        'spoofing': string.Template(
"""
Scenario: Spoofing
    Given $process causes data to flow from $source in $sourceZone to $destination in $destinationZone
    When $source attempts to impersonate something or someone else related to $process
    Then implement and validate digital identity by ...

# Mitigation details
# https://github.com/OWASP/ASVS/blob/master/4.0/en/0x11-V2-Authentication.md
# https://github.com/OWASP/ASVS/blob/master/4.0/en/0x12-V3-Session-management.md
# https://owasp.org/www-project-proactive-controls/v3/en/c6-digital-identity
"""
        ),
        'tampering': string.Template(
"""
Scenario: Tampering
    Given $process causes data to flow from less trusted $source in $sourceZone to more trusted $destination in $destinationZone
    When $source modifies or otherwise tampers with data related to $process
    Then treat all input as malcious and handle it safely by ...

# Mitigation details
# https://github.com/OWASP/ASVS/blob/master/4.0/en/0x18-V10-Malicious.md
"""        
        ),
        'repudiation': string.Template(
"""
Scenario: Repudiation
    Given $process has spoofing and tampering threats
    When $source claims to not have taken an action related to $process
    Then securely log all actions by ... 

# Mitigation details
# https://github.com/OWASP/ASVS/blob/master/4.0/en/0x15-V7-Error-Logging.md
"""        
        ),
        'informationDisclosure': string.Template(
"""
Scenario: Information Disclosure
    Given $process causes data to flow from a more trusted $source in $sourceZone to a less trusted $destination in $destinationZone
    When $source attempts to gain access to information it is not authorized to see related to $process
    Then ensure data is protected everywhere by ...

# Mitigation details
# https://github.com/OWASP/ASVS/blob/master/4.0/en/0x14-V6-Cryptography.md
# https://github.com/OWASP/ASVS/blob/master/4.0/en/0x16-V8-Data-Protection.md
"""
        ),
        'denialOfService': string.Template(
"""
Scenario: Denial of Service
    Given $process causes data to originate from $source outside of our control in $sourceZone
    When $source attempts to deny or degrade service related to $process
    Then ensure service resources are protected and all errors are handled by ...

# Mitigation details
# https://github.com/OWASP/ASVS/blob/master/4.0/en/0x19-V11-BusLogic.md
# https://github.com/OWASP/ASVS/blob/master/4.0/en/0x20-V12-Files-Resources.md
"""
        ),
        'elevationOfPrivilege': string.Template(
"""
Scenario: Elevation of Privilege
    Given $process causes data to flow from less trusted $source in $sourceZone to more trusted $destination in $destinationZone
    When $source attempts to gain addition capabilities without authorization related to $process
    Then ensure acces control is enforced by ...

# Mitigation details
# https://github.com/OWASP/ASVS/blob/master/4.0/en/0x12-V4-Access-Control.md
"""        
        )

    }




from ..shapes.Rect import Rect
from ..models.Node import Node 
from ..models.Text import Text 
from ..models.UserObject import UserObject 
from ..utils import MxUtils


class NodeFactory:
    def __init__(self, coords):
        super(NodeFactory, self).__init__()
        self.coords = coords

    def from_xml(self, xml, value):
        texts = []
        texts.append(Text.from_xml(xml.get('style')))
        
        if type(value) == str:
            label = value
        else:
            assert(MxUtils.is_user_object(value.xml))
            label = value.label
        
        return Node(
            xml=xml,
            sid=xml.get('id'),
            gid=xml.get('parent'),
            value=value,
            rect=Rect(
                x=int(xml.find('mxGeometry').get('x')),
                y=int(xml.find('mxGeometry').get('y')),
                width=int(xml.find('mxGeometry').get('width')),
                height=int(xml.find('mxGeometry').get('height'))
            ),
            label=label,
            texts=texts,
            fill=None,
            stroke=None
        )


from .GraphObj import GraphObj


class Node(GraphObj):
    def __init__(self, xml, sid, gid, value, label, rect, texts, fill, stroke):

        super(Node, self).__init__(sid, gid, value)

        self.xml = xml
        self.rect = rect
        self.texts = texts
        self.fill = fill
        self.stroke = stroke
        self.label = label
        self.shape = None

    def get_value(self, name):
        return self.xml.get(name)

    def set_value(self, name, value):
        return self.xml.set(name, value)
    
    def text_to_mx_value(self):
        value = ""
        last_text = len(self.texts) - 1
        for i, t in enumerate(self.texts):
            style = t.get_mx_style()
            value += "<p style='" + style + "'>" + t.text + "</p>"
            if i != last_text:
                value += "<hr size='1'/>"
        return value



import json, argparse, os
from .db import Node, Edge, dbgraph
from .mx.utils import MxUtils
from .gherkin_stride import create_gherkins_from_threats, create_feature_file_for_gherkins



class ThreatMaterializer(object):

    @classmethod
    def get_flows_with_threats(cls):

        SPOOFING = 'spoofing'
        TAMPERING = 'tampering'
        REPUDIATION = 'repudiation'
        INFORMATION_DISCLOSURE = 'informationDisclosure'
        DENIAL_OF_SERVICE = 'denialOfService'
        ELEVATION_OF_PRIVILEGE = 'elevationOfPrivilege'

        SOURCE = 'source'
        SOURCE_ZONE = 'sourceZone'
        DESTINATION = 'destination'
        DESTINATION_ZONE =  'destinationZone'
        PROCESS = 'process'

        threats = { 
            SPOOFING: [],
            TAMPERING: [],
            REPUDIATION: [],
            INFORMATION_DISCLOSURE: [],
            DENIAL_OF_SERVICE: [],
            ELEVATION_OF_PRIVILEGE: []
        }

        Source = Node.alias()
        Destination = Node.alias()
        Process = Node.alias()

        edgequery = (
            Edge.select(
                Source.label.alias(SOURCE), 
                Source.zone.alias(SOURCE_ZONE),
                Destination.label.alias(DESTINATION), 
                Destination.zone.alias(DESTINATION_ZONE),
                Process.label.alias(PROCESS)
            )
            .join(Source, on=(Source.id == Edge.source))
            .switch(Process)
            .join(Process, on=(Process.id == Edge.process))
            .switch(Destination)
            .join(Destination, on=(Destination.id == Edge.destination))
        ) 

        threats[ELEVATION_OF_PRIVILEGE] = list(
            edgequery.where(
                (Source.zone < Destination.zone)
            ).dicts()
        )

        threats[SPOOFING] = list(
        edgequery.where(
                (Source.zone == 0) & 
                (Destination.zone == 1)
            ).dicts()
        )

        threats[TAMPERING] = list(
            edgequery.where(
                (Source.zone < Destination.zone)
            ).dicts()
        )

        threats[REPUDIATION] = [threat for threat in threats[SPOOFING] if threat in threats[TAMPERING]]

        threats[DENIAL_OF_SERVICE] = list(
            edgequery.where(
                (Source.zone == 0) & 
                (Destination.zone == 1)
            ).dicts()
        )

        threats[INFORMATION_DISCLOSURE] = list(
            edgequery.where(
                (Source.zone > Destination.zone)
            ).dicts()
        )

        return threats

    @classmethod
    def output_threats(cls, threats):
        print(
            json.dumps(threats, indent=4)
        )

    @classmethod
    def materialize(cls):

        args = argparse.ArgumentParser(description="Enumerate STRIDE threats from a data flow diagram and create test case stubs")
        args.add_argument(
            "--diagram", 
            default="samples/sample.drawio",
            type=argparse.FileType('r'), 
            help="The draw.io data flow diagram filename"
        )
        filename = args.parse_args().diagram.name
        
        args.add_argument(
            "--featurefile",
            default=os.path.basename(filename) + ".feature",
            type=argparse.FileType('w+'),
            help="The feature filename to write"
        )

        graph = MxUtils.parse_from_xml(file=args.parse_args().diagram)
        zones = dbgraph.get_node_trust_zones_from_graph(graph)
        dbgraph.load_graph_into_db(graph, zones)

        threats = cls.get_flows_with_threats()

        gherkin_candidates = create_gherkins_from_threats(threats)
        feature_file = create_feature_file_for_gherkins(feature=filename, gherkins=gherkin_candidates)

        args.parse_args().featurefile.write(feature_file)
        
        cls.output_threats(threats)  




import string
from gherkin.token_scanner import TokenScanner
from gherkin.token_matcher import TokenMatcher
from gherkin.parser import Parser
from gherkin.errors import ParserError, CompositeParserException

from .templates import Base, Scenarios

def create_gherkins_from_threats(threats):
    scenarios = Scenarios.stride

    gherkins = list()
   
    for threat_class in threats:
        for threat in threats[threat_class]:

            threat_gherkin = scenarios[threat_class].substitute(
                process = threat['process'],
                source = threat['source'],
                sourceZone = threat['sourceZone'],
                destination = threat['destination'],
                destinationZone = threat['destinationZone']
            )

            parser = Parser()
            feature_base = Base.feature_base

            try:
                parser.parse(
                    TokenScanner(
                        feature_base.substitute(
                            component="None",
                            scenario=threat_gherkin
                        )
                    )
                )
            except CompositeParserException:
                print("Invalid gherkin template created: {}".format(threat_gherkin))
            else:
                gherkins.append(threat_gherkin)
    
    return gherkins

def create_feature_file_for_gherkins(feature, gherkins):
    feature_file = Base.feature_bare.substitute(component=feature)
    for gherkin in gherkins:
        feature_file += gherkin

    Parser().parse(
        TokenScanner(feature_file)
    )
    return feature_file


from ..utils import MxConst
from .GraphObj import GraphObj


class Edge(GraphObj):
    def __init__(self, sid, gid, value, fr, to, curve):
       
        super(Edge, self).__init__(sid, gid, value)

        self.fr = fr
        self.to = to
        self.curve = curve
        self.style = None
        self.dir = None
        self.arrowtail = None

    def curve_start_end(self):
        if self.dir == MxConst.BACK:
            return self.curve.end, self.curve.start
        else:
            return self.curve.start, self.curve.end



ELLIPSE = "ellipse"
OVAL = "oval"
BOX = "box"
RECT = "rect"
RECTANGLE = "rectangle"
HEXAGON = "hexagon"
POLYGON = "polygon"
CIRCLE = "circle"
EGG = "egg"
TRIANGLE = "triangle"
PLAIN = "plain"
DIAMOND = "diamond"
TRAPEZIUM = "trapezium"
PARALLELOGRAM = "parallelogram"
HOUSE = "house"
PENTAGON = "pentagon"
OCTAGON = "octagon"
DOUBLE_CIRCLE = "doublecircle"
DOUBLE_OCTAGON = "doubleoctagon"
INV_TRIANGLE = "invtriangle"
INV_TRAPEZIUM = "invtrapezium"
INV_HOUSE = "invhouse"
SQUARE = "square"
STAR = "star"
UNDERLINE = "underline"
CYLINDER = "cylinder"
NOTE = "note"
TAB = "tab"
FOLDER = "folder"
BOX_3D = "box3d"
COMPONENT = "component"
PROMOTER = "promoter"
RPROMOTER = "rpromoter"
LPROMOTER = "lpromoter"
CDS = "cds"
RARROW = "rarrow"
LARROW = "larrow"



from abc import ABC


class GraphObj(ABC):
    def __init__(self, sid, gid, value):
        self.sid = sid
        self.gid = gid
        self.value = value


    def enrich_from_graph(self, attrs):
        for e in attrs:
            self.__setattr__(e[0], e[1])



# 
# Simple DEFLATE decompressor
# Copyright (c) Project Nayuki
# 
# https://www.nayuki.io/page/simple-deflate-decompressor
# https://github.com/nayuki/Simple-DEFLATE-decompressor
# 

import sys
python3 = sys.version_info.major >= 3
if python3:
	import io
else:
	import StringIO


class CanonicalCode(object):
	
	"""A canonical Huffman code, where the code values for each symbol is derived
	from a given sequence of code lengths. This data structure is immutable.
	This could be transformed into an explicit Huffman code tree.
	
	Example:
	  Code lengths (canonical code):
	    Symbol A: 1
	    Symbol B: 0 (no code)
	    Symbol C: 3
	    Symbol D: 2
	    Symbol E: 3
	  
	  Generated Huffman codes:
	    Symbol A: 0
	    Symbol B: (Absent)
	    Symbol C: 110
	    Symbol D: 10
	    Symbol E: 111
	  
	  Huffman code tree:
	      .
	     / \
	    A   .
	       / \
	      D   .
	         / \
	        C   E"""
	
	def __init__(self, codelengths):
		"""Creates a canonical Huffman code from the given list of symbol code lengths.
		Each code length must be non-negative. Code length 0 means no code for the symbol.
		The collection of code lengths must represent a proper full Huffman code tree.
		Examples of code lengths that result in correct full Huffman code trees:
		- [1, 1] (result: A=0, B=1)
		- [2, 2, 1, 0, 0, 0] (result: A=10, B=11, C=0)
		- [3, 3, 3, 3, 3, 3, 3, 3] (result: A=000, B=001, C=010, ..., H=111)
		Examples of code lengths that result in under-full Huffman code trees:
		- [0, 2, 0] (result: B=00, unused=01, unused=1)
		- [0, 1, 0, 2] (result: B=0, D=10, unused=11)
		Examples of code lengths that result in over-full Huffman code trees:
		- [1, 1, 1] (result: A=0, B=1, C=overflow)
		- [1, 1, 2, 2, 3, 3, 3, 3] (result: A=0, B=1, C=overflow, ...)"""
		
		# Check argument values
		if any(x < 0 for x in codelengths):
			raise ValueError("Negative code length")
		
		# This dictionary maps Huffman codes to symbol values.
		# Each key is the Huffman code padded with a 1 bit at the
		# beginning to disambiguate codes of different lengths
		# (e.g. otherwise we can't distinguish 0b01 from 0b0001).
		# For the example of codelengths=[1,0,3,2,3], we would have:
		#     0b1_0 -> 0
		#    0b1_10 -> 3
		#   0b1_110 -> 2
		#   0b1_111 -> 4
		self._code_bits_to_symbol = {}
		
		# Allocate code values to symbols. Symbols are processed in the order
		# of shortest code length first, breaking ties by lowest symbol value.
		nextcode = 0
		for codelength in range(1, max(codelengths) + 1):
			nextcode <<= 1
			startbit = 1 << codelength
			for (symbol, cl) in enumerate(codelengths):
				if cl != codelength:
					continue
				if nextcode >= startbit:
					raise ValueError("This canonical code produces an over-full Huffman code tree")
				self._code_bits_to_symbol[startbit | nextcode] = symbol
				nextcode += 1
		if nextcode != 1 << max(codelengths):
			raise ValueError("This canonical code produces an under-full Huffman code tree")
	
	
	def decode_next_symbol(self, inp):
		"""Decodes the next symbol from the given bit input stream based on this
		canonical code. The returned symbol value is in the range [0, len(codelengths))."""
		codebits = 1  # The start bit
		while True:
			# Accumulate one bit at a time on the right side until a match is found
			# in the code_bits_to_symbol dictionary. Because the Huffman code tree is
			# full, this loop must terminate after at most max(codelengths) iterations.
			codebits = codebits << 1 | inp.read_no_eof()
			result = self._code_bits_to_symbol.get(codebits, -1)
			if result != -1:
				return result
	
	
	def __str__(self):
		"""Returns a string representation of this canonical code,
		useful for debugging only, and the format is subject to change."""
		return "\n".join(
			"Code {}: Symbol {}".format(bin(codebits)[3 : ], symbol)
			for (codebits, symbol) in sorted(self._code_bits_to_symbol.items())) + "\n"



class Decompressor(object):
	
	# -- Public functions --
	
	"""Decompresses raw DEFLATE data (without zlib or gzip container) into bytes."""
	
	@staticmethod
	def decompress_to_bytes(bitin):
		"""Reads from the given input stream, decompress the data, and returns a new byte list."""
		out = io.BytesIO() if python3 else StringIO.StringIO()
		Decompressor.decompress_to_stream(bitin, out)
		return out.getvalue()
	
	
	@staticmethod
	def decompress_to_stream(bitin, out):
		"""Reads from the given input stream, decompress the data, and writes to the given output stream."""
		Decompressor(bitin, out)
	
	
	# -- Private implementation --
	
	def __init__(self, bitin, out):
		# Initialize fields
		self._input = bitin
		self._output = out
		self._dictionary = ByteHistory(32 * 1024)
		
		# Process the stream of blocks
		while True:
			# Read the block header
			isfinal = bitin.read_no_eof() == 1  # bfinal
			type = self._read_int(2)  # btype
			
			# Decompress rest of block based on the type
			if type == 0:
				self._decompress_uncompressed_block()
			elif type == 1:
				self._decompress_huffman_block(Decompressor.FIXED_LITERAL_LENGTH_CODE, Decompressor.FIXED_DISTANCE_CODE)
			elif type == 2:
				litlencode, distcode = self._decode_huffman_codes()
				self._decompress_huffman_block(litlencode, distcode)
			elif type == 3:
				raise ValueError("Reserved block type")
			else:
				assert False, "Impossible value"
			if isfinal:
				break
	
	
	# -- The constant code trees for static Huffman codes (btype = 1) --
	
	FIXED_LITERAL_LENGTH_CODE = CanonicalCode([8]*144 + [9]*112 + [7]*24 + [8]*8)
	
	FIXED_DISTANCE_CODE = CanonicalCode([5] * 32)
	
	
	# -- Method for reading and decoding dynamic Huffman codes (btype = 2) --
	
	# Reads from the bit input stream, decodes the Huffman code
	# specifications into code trees, and returns the trees.
	def _decode_huffman_codes(self):
		numlitlencodes = self._read_int(5) + 257  # hlit + 257
		numdistcodes = self._read_int(5) + 1      # hdist + 1
		
		numcodelencodes = self._read_int(4) + 4   # hclen + 4
		codelencodelen = [0] * 19  # This list is filled in a strange order
		codelencodelen[16] = self._read_int(3)
		codelencodelen[17] = self._read_int(3)
		codelencodelen[18] = self._read_int(3)
		codelencodelen[ 0] = self._read_int(3)
		for i in range(numcodelencodes - 4):
			j = (8 + i // 2) if (i % 2 == 0) else (7 - i // 2)
			codelencodelen[j] = self._read_int(3)
		
		# Create the code length code
		codelencode = CanonicalCode(codelencodelen)
		
		# Read the main code lengths and handle runs
		codelens = []
		while len(codelens) < numlitlencodes + numdistcodes:
			sym = codelencode.decode_next_symbol(self._input)
			if 0 <= sym <= 15:
				codelens.append(sym)
			elif sym == 16:
				if len(codelens) == 0:
					raise ValueError("No code length value to copy")
				runlen = self._read_int(2) + 3
				codelens.extend(codelens[-1 : ] * runlen)
			elif sym == 17:
				runlen = self._read_int(3) + 3
				codelens.extend([0] * runlen)
			elif sym == 18:
				runlen = self._read_int(7) + 11
				codelens.extend([0] * runlen)
			else:
				assert False, "Symbol out of range"
		if len(codelens) > numlitlencodes + numdistcodes:
			raise ValueError("Run exceeds number of codes")
		
		# Create literal-length code tree
		litlencode = CanonicalCode(codelens[ : numlitlencodes])
		
		# Create distance code tree with some extra processing
		distcodelen = codelens[numlitlencodes : ]
		if len(distcodelen) == 1 and distcodelen[0] == 0:
			distcode = None  # Empty distance code; the block shall be all literal symbols
		else:
			# Get statistics for upcoming logic
			onecount = sum(1 for x in distcodelen if x == 1)
			otherpositivecount = sum(1 for x in distcodelen if x > 1)
			
			# Handle the case where only one distance code is defined
			if onecount == 1 and otherpositivecount == 0:
				# Add a dummy invalid code to make the Huffman tree complete
				distcodelen.extend([0] * (32 - len(distcodelen)))
				distcodelen[31] = 1
			distcode = CanonicalCode(distcodelen)
		
		return (litlencode, distcode)
	
	
	# -- Block decompression methods --
	
	# Handles and copies an uncompressed block from the bit input stream.
	def _decompress_uncompressed_block(self):
		# Discard bits to align to byte boundary
		while self._input.get_bit_position() != 0:
			self._input.read_no_eof()
		
		# Read length
		len  = self._read_int(16)
		nlen = self._read_int(16)
		if len ^ 0xFFFF != nlen:
			raise valueError("Invalid length in uncompressed block")
		
		# Copy bytes
		for _ in range(len):
			b = self._input.read_byte()
			if b == -1:
				raise EOFError()
			self._output.write(bytes((b,)) if python3 else chr(b))
			self._dictionary.append(b)
	
	
	# Decompresses a Huffman-coded block from the bit input stream based on the given Huffman codes.
	def _decompress_huffman_block(self, litlencode, distcode):
		# litlencode cannot be None, but distcode is allowed to be None
		while True:
			sym = litlencode.decode_next_symbol(self._input)
			if sym == 256:  # End of block
				break
			
			if sym < 256:  # Literal byte
				self._output.write(bytes((sym,)) if python3 else chr(sym))
				self._dictionary.append(sym)
			else:  # Length and distance for copying
				run = self.decode_run_length(sym)
				assert 3 <= run <= 258, "Invalid run length"
				if distcode is None:
					raise ValueError("Length symbol encountered with empty distance code")
				distsym = distcode.decode_next_symbol(self._input)
				dist = self._decode_distance(distsym)
				assert 1 <= dist <= 32768, "Invalid distance"
				self._dictionary.copy(dist, run, self._output)
	
	
	# -- Symbol decoding methods --
	
	# Returns the run length based on the given symbol and possibly reading more bits.
	def decode_run_length(self, sym):
		# Symbols outside the range cannot occur in the bit stream;
		# they would indicate that the decompressor is buggy
		assert 257 <= sym <= 287, "Invalid run length symbol: " + str(sym)
		
		if sym <= 264:
			return sym - 254
		elif sym <= 284:
			numextrabits = (sym - 261) // 4
			return (((sym - 265) % 4 + 4) << numextrabits) + 3 + self._read_int(numextrabits)
		elif sym == 285:
			return 258
		else:  # sym is 286 or 287
			raise ValueError("Reserved length symbol: " + str(sym))
	
	
	# Returns the distance based on the given symbol and possibly reading more bits.
	def _decode_distance(self, sym):
		# Symbols outside the range cannot occur in the bit stream;
		# they would indicate that the decompressor is buggy
		assert 0 <= sym <= 31, "Invalid distance symbol: " + str(sym)
		
		if sym <= 3:
			return sym + 1
		elif sym <= 29:
			numextrabits = sym // 2 - 1
			return ((sym % 2 + 2) << numextrabits) + 1 + self._read_int(numextrabits)
		else:  # sym is 30 or 31
			raise ValueError("Reserved distance symbol: " + str(sym))
	
	
	# -- Utility method --
	
	# Reads the given number of bits from the bit input stream as a single integer, packed in little endian.
	def _read_int(self, numbits):
		if numbits < 0:
			raise ValueError()
		return sum(self._input.read_no_eof() << i for i in range(numbits))



class ByteHistory(object):
	
	"""Stores a finite recent history of a byte stream. Useful as an implicit
	dictionary for Lempel-Ziv schemes. Mutable and not thread-safe."""
	
	def __init__(self, size):
		"""Creates a byte history of the given size, initialized to zeros."""
		if size < 1:
			raise ValueError("Size must be positive")
		# Circular buffer of byte data.
		self._data = [0] * size
		# Index of next byte to write to, always in the range [0, len(data)).
		self._index = 0
	
	
	def append(self, b):
		"""Appends the specified byte to this history.
		This overwrites the byte value at 'size' positions ago."""
		assert 0 <= self._index < len(self._data)
		self._data[self._index] = b
		self._index = (self._index + 1) % len(self._data)
	
	
	def copy(self, dist, count, out):
		"""Copies 'count' bytes starting at 'dist' bytes ago to the
		given output stream and also back into this buffer itself.
		Note that if the count exceeds the distance, then some of the output
		data will be a copy of data that was copied earlier in the process."""
		if count < 0 or not (1 <= dist <= len(self._data)):
			raise ValueError()
		
		readindex = (self._index - dist) % len(self._data)
		for _ in range(count):
			b = self._data[readindex]
			readindex = (readindex + 1) % len(self._data)
			out.write(bytes((b,)) if python3 else chr(b))
			self.append(b)



class BitInputStream(object):
	
	"""A stream of bits that can be read. Because they come from an underlying byte stream, the
	total number of bits is always a multiple of 8. Bits are packed in little endian within a byte.
	For example, the byte 0x87 reads as the sequence of bits [1,1,1,0,0,0,0,1]."""
	
	def __init__(self, inp):
		"""Constructs a bit input stream based on the given byte input stream."""
		# The underlying byte stream to read from.
		self._input = inp
		# Either in the range [0x00, 0xFF] if bits are available, or -1 if end of stream is reached.
		self._current_byte = 0
		# Number of remaining bits in the current byte, always between 0 and 7 (inclusive).
		self._num_bits_remaining = 0
	
	
	def get_bit_position(self):
		"""Returns the current bit position, which ascends from 0 to 7 as bits are read."""
		assert 0 <= self._num_bits_remaining <= 7
		return -self._num_bits_remaining % 8
	
	
	def read_byte(self):
		"""Discards the remainder of the current byte (if any) and reads the next
		whole byte from the stream. Returns -1 if the end of stream is reached."""
		self._current_byte = 0
		self._num_bits_remaining = 0
		b = self._input.read(1)
		if len(b) == 0:
			return -1
		return b[0] if python3 else ord(b)
	
	
	def read(self):
		"""Reads a bit from this stream. Returns 0 or 1 if a bit is available, or -1 if
		the end of stream is reached. The end of stream always occurs on a byte boundary."""
		if self._current_byte == -1:
			return -1
		if self._num_bits_remaining == 0:
			b = self._input.read(1)
			if len(b) == 0:
				self._current_byte = -1
				return -1
			self._current_byte = b[0] if python3 else ord(b)
			self._num_bits_remaining = 8
		assert self._num_bits_remaining > 0
		self._num_bits_remaining -= 1
		return (self._current_byte >> (7 - self._num_bits_remaining)) & 1
	
	
	def read_no_eof(self):
		"""Reads a bit from this stream. Returns 0 or 1 if a bit is available, or raises an EOFError
		if the end of stream is reached. The end of stream always occurs on a byte boundary."""
		result = self.read()
		if result == -1:
			raise EOFError()
		return result
	
	
	def close(self):
		"""Closes this stream and the underlying input stream."""
		self._input.close()
		self._current_byte = -1
		self._num_bits_remaining = 0



from ..models.UserObject import UserObject


class UserObjectFactory:
    ZONE_OBJ = 'trust zone'

    def __init__(self):
        super(UserObjectFactory, self).__init__()

    def from_xml(self, xml):
        #uo_type = xml.get('type')


        """
        TODO: 
        Deprecate zone objects, just let them be user objects
        """
        #if uo_type == self.ZONE_OBJ:
        #   return ZoneObject(xml=xml, label=xml.get('label'))
        #else:
        return UserObject(xml=xml, label=xml.get('label'))

    def from_node(self, node):
        pass


from ..utils import MxConst
from .Styles import Styles


class Text:
    def __init__(self, anchor, family, size, text, color):
        self.anchor = anchor
        self.family = family
        self.size = size
        self.text = text
        self.color = color

    def get_mx_style(self):
        align = MxConst.CENTER if self.anchor == MxConst.MIDDLE else MxConst.START
        # TODO: add right
        margin = (
            "margin-top:4px;" if self.anchor == MxConst.MIDDLE else "margin-left:4px;"
        )
        rescaled_size = 10.0 * (self.size / 14.0)
        return Styles.TEXT.format(
            align=align,
            margin=margin,
            size=rescaled_size,
            family=self.family or MxConst.DEFAULT_FONT_FAMILY,
            color=self.color or MxConst.DEFAULT_FONT_COLOR,
        )

    @staticmethod
    def from_xml(text):
        if text != None:
            text =  text.replace("<", "&lt;").replace(">", "&gt;")
        else: 
            text = ""

        return Text(
            text = text,
            anchor=None,
            family=MxConst.DEFAULT_FONT_FAMILY,
            size=MxConst.DEFAULT_TEXT_SIZE,
            color=None,
        )



class CoordsTranslate:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def complex_translate(self, cnum):
        return complex(cnum.real + self.x, cnum.imag + self.y)

    def translate(self, x, y):
        return float(x) + self.x, float(y) + self.y

    @staticmethod
    def from_xml_transform(xml):
        x = xml.get('dx')
        y = xml.get('dy')
        return CoordsTranslate(x=float(x), y=float(y))



from argparse import ArgumentParser


class Arguments(ArgumentParser):
    def __init__(self, version):
        super(Arguments, self).__init__(prog="graphviz2drawio")
        self.add_argument("to_convert", metavar="file.dot", help="graphviz file")
        output = self.add_mutually_exclusive_group()
        output.add_argument("outfile", help="output file", nargs="?", default=None)
        output.add_argument("--stdout", action="store_true", help="print xml to stdout")
        self.add_argument(
            "--version",
            action="version",
            version="%(prog)s {version}".format(version=version),
        )
        self.add_argument(
            "-p",
            "--program",
            help="layout program (defaults to dot), may be one of: neato, dot, twopi, circo, fdp, nop",
            default="dot",
        )



class Rect:
    def __init__(self, x, y, width, height):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.top = y
        self.bottom = y + height
        self.left = x
        self.right = x + width

    def is_overlapping(self, inner):

        # TODO: make this handle any overlap case, inner or outer
        """      X
         0 1 2 3 4 
        -1    -----  <-- top
        -2    |   |
        -3    |   | <-- right
      Y -4    ----- <-- bottom
        """
        if self.top > inner.bottom or self.bottom < inner.top:
            return False
        if self.right < inner.left or self.left > inner.left:
            return False

        return True

    def x_ratio(self, search):
        if search < self.x:
            return 0
        elif search > self.x + self.width:
            return 1
        else:
            ratio = (search - self.x) / self.width
            return self._approx(ratio, 0.5, 0.1)

    def y_ratio(self, search):
        if search < self.y:
            return 0
        elif search > self.y + self.height:
            return 1
        else:
            ratio = (search - self.y) / self.height
            return self._approx(ratio, 0.5, 0.1)

    @staticmethod
    def _approx(value, center, delta):
        if abs(value - center) < delta:
            return center
        return value

    def to_dict_str(self):
        return {
            "x": str(self.x),
            "y": str(self.y),
            "width": str(self.width),
            "height": str(self.height),
        }



import xml.etree.ElementTree, io, base64, urllib.parse, collections

from ..models import MxGraph, Edge, Node

from ..io.NodeFactory import NodeFactory
from ..io.UserObjectFactory import UserObjectFactory
from ..io.EdgeFactory import EdgeFactory

from . import MxConst
from . import deflatedecompress

def is_edge(element):
    return element.get('edge') == str(1)

def is_user_object(element):
    return element.tag == MxConst.USER_OBJECT

def decode_xml_element_to_cell(element, coords):
    edgefactory = EdgeFactory(coords)
    nodefactory = NodeFactory(coords)
    userobjectfactory = UserObjectFactory()

    value = element.get('value')

    if is_edge(element):
        cell = edgefactory.from_xml(xml=element, value=None)
        return cell
        #return edgefactory.from_xml(xml=element, value=None)

    if is_user_object(element):
        value = userobjectfactory.from_xml(element)
        
        id = value.xml.get('id')
        label = value.xml.get('label')
    
        if value is not None:
            # extract the plain node it wraps, invert the relationship
            # old: userobject.node
            element = value.xml.find(MxConst.CELL)
            element.set('id', id)
            element.set('label', label)

    cell = nodefactory.from_xml(xml=element, value=value)
    return cell


def get_mxgraph_from_xml(root, elements):
    from ..shapes.CoordsTranslate import CoordsTranslate
    from collections import OrderedDict

    coords = CoordsTranslate.from_xml_transform(root)

    edges = []
    nodes = collections.OrderedDict()

    for element in elements:
        # Every mxGraph contains an existing element of id 0 and 1, skip them
        if element.get('id') == str(0) or element.get('id') == str(1):
            continue
        
        cell = decode_xml_element_to_cell(element, coords)

        if type(cell) == Edge.Edge:
            edges.append(cell)
        else:
            nodes[cell.sid] = cell

    return(MxGraph.MxGraph(nodes=nodes, edges=edges))


def decompress_diagram_to_file(compressed_diagram):
    compressed_bytes = io.BytesIO(
        base64.b64decode(compressed_diagram)
    )

    urlencoded_diagram = deflatedecompress.Decompressor.decompress_to_bytes(
        deflatedecompress.BitInputStream(compressed_bytes)
    )

    diagram_xml = urllib.parse.unquote(
        urlencoded_diagram.decode('utf-8')
    )

    return(
        io.BytesIO(
            bytes(diagram_xml, 'utf-8')
        )
    )

def parse_from_xml(file):

    tree = xml.etree.ElementTree.parse(file)
    root = tree.getroot()

    diagram = root.find(MxConst.DIAGRAM)

    tree = xml.etree.ElementTree.parse(
        decompress_diagram_to_file(diagram.text)
    )

    diagram_root = tree.getroot()
    cells = diagram_root.find(MxConst.ROOT).getchildren()
    
    return get_mxgraph_from_xml(diagram_root, cells)


from enum import Enum
from . import Shape


class Styles(Enum):
    NODE = "verticalAlign=top;align=left;overflow=fill;html=1;rounded=0;shadow=0;comic=0;labelBackgroundColor=none;strokeColor={stroke};strokeWidth=1;fillColor={fill};"
    EDGE = "rounded=1;html=1;exitX={exit_x:.3g};exitY={exit_y:.3g};entryX={entry_x:.3g};entryY={entry_y:.3g};jettySize=auto;orthogonalLoop=1;endArrow={end_arrow};dashed={dashed};endFill={end_fill};"
    TEXT = "margin:0px;text-align:{align};{margin};font-size:{size}px;font-family:{family};color:{color};"

    ELLIPSE = "ellipse;" + NODE
    CIRCLE = "ellipse;aspect=fixed;" + NODE
    HEXAGON = "shape=hexagon;perimeter=hexagonPerimeter2;" + NODE
    EGG = "shape=mxgraph.flowchart.display;direction=south;" + NODE
    TRIANGLE = "triangle;direction=north;" + NODE
    LINE = (
        "line;strokeWidth=2;verticalAlign=bottom;labelPosition=center;verticalLabelPosition=top;align=center;"
        + NODE
    )
    DIAMOND = "rhombus;" + NODE
    TRAPEZOID = "shape=trapezoid;perimeter=trapezoidPerimeter;" + NODE
    PARALLELOGRAM = "shape=parallelogram;perimeter=parallelogramPerimeter;" + NODE
    HOUSE = "shape=offPageConnector;direction=west;" + NODE
    PENTAGON = "shape=mxgraph.basic.pentagon;" + NODE
    OCTAGON = (
        "shape=mxgraph.basic.octagon2;align=center;verticalAlign=middle;dx=15;" + NODE
    )
    DOUBLE_CIRCLE = "ellipse;shape=doubleEllipse;aspect=fixed;" + NODE
    DOUBLE_OCTAGON = (
        "shape=image;html=1;verticalAlign=middle;verticalLabelPosition=middle;imageAspect=0;aspect=fixed;image=https://cdn4.iconfinder.com/data/icons/feather/24/octagon-128.png;labelPosition=center;align=center;"
        + NODE
    )
    INV_TRIANGLE = "triangle;direction=south;" + NODE
    INV_TRAPEZOID = (
        "shape=trapezoid;perimeter=trapezoidPerimeter;direction=west;" + NODE
    )
    INV_HOUSE = "shape=offPageConnector;direction=east;" + NODE
    SQUARE = "aspect=fixed;" + NODE
    STAR = (
        "shape=mxgraph.basic.star;labelPosition=center;align=center;verticalLabelPosition=middle;verticalAlign=middle;"
        + NODE
    )
    UNDERLINE = "line;strokeWidth=2;verticalAlign=bottom;labelPosition=center;verticalLabelPosition=top;align=center;"
    CYLINDER = "shape=cylinder;boundedLbl=1;backgroundOutline=1;" + NODE
    NOTE = "shape=note;backgroundOutline=1;" + NODE
    TAB = "shape=folder;tabWidth=40;tabHeight=14;tabPosition=left;" + NODE
    FOLDER = (
        "shape=mxgraph.office.concepts.folder;outlineConnect=0;align=center;verticalLabelPosition=middle;verticalAlign=middle;labelPosition=center;shadow=0;dashed=0;"
        + NODE
    )
    CUBE = "shape=cube;boundedLbl=1;backgroundOutline=1;" + NODE
    COMPONENT = (
        "shape=component;align=center;spacingLeft=36;verticalAlign=bottom;" + NODE
    )
    RPROMOTER = (
        "shape=mxgraph.arrows2.bendArrow;dy=15;dx=38;notch=0;arrowHead=55;rounded=0;shadow=0;dashed=0;align=center;verticalAlign=middle;"
        + NODE
    )
    LPROMOTER = "flipH=1;" + RPROMOTER
    CDS = (
        "shape=mxgraph.arrows2.arrow;dy=0;dx=10;notch=0;shadow=0;dashed=0;align=center;verticalAlign=middle;"
        + NODE
    )
    RARROW = (
        "shape=mxgraph.arrows2.arrow;dy=0.6;dx=40;align=center;labelPosition=center;notch=0;strokeWidth=2;verticalLabelPosition=middle;verticalAlign=middle;"
        + NODE
    )
    LARROW = "flipH=1;" + RARROW

    @staticmethod
    def get_for_shape(dot_shape):
        if dot_shape in (None, Shape.ELLIPSE, Shape.OVAL):
            return Styles.ELLIPSE
        elif dot_shape in (Shape.BOX, Shape.RECT, Shape.RECTANGLE):
            return Styles.NODE
        elif dot_shape in (Shape.HEXAGON, Shape.POLYGON):
            return Styles.HEXAGON
        elif dot_shape == Shape.CIRCLE:
            return Styles.CIRCLE
        elif dot_shape == Shape.EGG:
            return Styles.EGG
        elif dot_shape == Shape.TRIANGLE:
            return Styles.TRIANGLE
        elif dot_shape == Shape.PLAIN:
            return Styles.LINE
        elif dot_shape == Shape.DIAMOND:
            return Styles.DIAMOND
        elif dot_shape == Shape.TRAPEZIUM:
            return Styles.TRAPEZOID
        elif dot_shape == Shape.PARALLELOGRAM:
            return Styles.PARALLELOGRAM
        elif dot_shape == Shape.HOUSE:
            return Styles.HOUSE
        elif dot_shape == Shape.PENTAGON:
            return Styles.PENTAGON
        elif dot_shape == Shape.OCTAGON:
            return Styles.OCTAGON
        elif dot_shape == Shape.DOUBLE_CIRCLE:
            return Styles.DOUBLE_CIRCLE
        elif dot_shape == Shape.DOUBLE_OCTAGON:
            return Styles.DOUBLE_OCTAGON
        elif dot_shape == Shape.INV_TRIANGLE:
            return Styles.INV_TRIANGLE
        elif dot_shape == Shape.INV_TRAPEZIUM:
            return Styles.INV_TRAPEZOID
        elif dot_shape == Shape.INV_HOUSE:
            return Styles.INV_HOUSE
        elif dot_shape == Shape.SQUARE:
            return Styles.SQUARE
        elif dot_shape == Shape.STAR:
            return Styles.STAR
        elif dot_shape == Shape.UNDERLINE:
            return Styles.UNDERLINE
        elif dot_shape == Shape.CYLINDER:
            return Styles.CYLINDER
        elif dot_shape == Shape.NOTE:
            return Styles.NODE
        elif dot_shape == Shape.TAB:
            return Styles.TAB
        elif dot_shape == Shape.FOLDER:
            return Styles.FOLDER
        elif dot_shape == Shape.BOX_3D:
            return Styles.CUBE
        elif dot_shape == Shape.COMPONENT:
            return Styles.COMPONENT
        elif dot_shape in (Shape.PROMOTER, Shape.RPROMOTER):
            return Styles.RPROMOTER
        elif dot_shape == Shape.LPROMOTER:
            return Styles.LPROMOTER
        elif dot_shape == Shape.CDS:
            return Styles.CDS
        elif dot_shape == Shape.RARROW:
            return Styles.RARROW
        elif dot_shape == Shape.LARROW:
            return Styles.LARROW
        else:
            return Styles.NODE

    def format(self, **values):
        return self.value.format(**values)



DECLARATION = '<?xml version="1.0"?>'
BACK = "back"
NO_FILL = "o"

SOLID = "solid"
DASHED = "dashed"

MIDDLE = "middle"

BLOCK = "block"
DIAMOND = "diamond"

CENTER = "center"
START = "start"

DEFAULT_TEXT_SIZE = 10
DEFAULT_FONT_FAMILY = "Helvetica"
DEFAULT_FONT_COLOR = "#000000"

DEFAULT_STROKE = "#000000"
DEFAULT_FILL = "#ffffff"

ARRAY = "Array"
CELL = "mxCell"
USER_OBJECT = "UserObject"
GEO = "mxGeometry"
GRAPH = "mxGraphModel"
POINT = "mxPoint"
ROOT = "root"
DIAGRAM = "diagram"

CURVE_INTERVALS = [0.25, 0.5, 0.75]

NONE = "none"



from xml.etree import ElementTree as ET

from ..utils import MxConst
from .Styles import Styles


class MxGraph:
    def __init__(self, nodes, edges):
        self.nodes = nodes
        self.edges = edges
        self.graph = ET.Element(MxConst.GRAPH)
        self.root = ET.SubElement(self.graph, MxConst.ROOT)
        ET.SubElement(self.root, MxConst.CELL, id="0")
        ET.SubElement(self.root, MxConst.CELL, id="1", parent="0")

        for edge in edges:
            self.add_edge(edge)
        for node in nodes.values():
            self.add_node(node)

    def add_edge(self, edge):
        source, target = self.get_edge_source_target(edge)
        style = self.get_edge_style(edge, source, target)

        edge_element = ET.SubElement(
            self.root,
            MxConst.CELL,
            id=edge.sid,
            style=style,
            parent="1",
            edge="1",
            source=source.sid,
            target=target.sid,
        )
        if edge.curve.cb is None:
            self.add_mx_geo(edge_element)
        else:
            self.add_mx_geo_with_points(edge_element, edge.curve)

    def get_edge_by_sid(self, sid):
        for edge in self.edges:
            if edge.sid == sid:
                return edge
        
        return None

    def get_edge_source_target(self, edge):
        if edge.dir == MxConst.BACK:
            return self.nodes[edge.to], self.nodes[edge.fr]
        else:
            return self.nodes[edge.fr], self.nodes[edge.to]

    def get_edge_style(self, edge, source_node, target_node):
        end_arrow = MxConst.BLOCK
        end_fill = 1
        dashed = 1 if edge.style == MxConst.DASHED else 0
        if edge.arrowtail is not None:
            tail = edge.arrowtail
            if edge.arrowtail[0] == MxConst.NO_FILL:
                end_fill = 0
                tail = edge.arrowtail[1:]
            if tail == MxConst.DIAMOND:
                end_arrow = MxConst.DIAMOND



        if edge.curve.cb is not None:
            style = "edgeStyle=orthogonalEdgeStyle;curved=1;" + style
        else:
            start_curve, end_curve = edge.curve_start_end()

            style = Styles.EDGE.format(
                entry_x=target_node.rect.x_ratio(end_curve.real),
                entry_y=target_node.rect.y_ratio(end_curve.imag),
                exit_x=source_node.rect.x_ratio(start_curve.real),
                exit_y=source_node.rect.y_ratio(start_curve.imag),
                end_arrow=end_arrow,
                dashed=dashed,
                end_fill=end_fill,
            )

        return style

    def add_node(self, node):
        fill = (
            node.fill
            if (node.fill is not None and node.fill != "none")
            else MxConst.DEFAULT_FILL
        )
        stroke = node.stroke if node.stroke is not None else MxConst.DEFAULT_STROKE

        style = Styles.get_for_shape(node.shape).format(fill=fill, stroke=stroke)

        node_element = ET.SubElement(
            self.root,
            MxConst.CELL,
            id=node.sid,
            value=node.text_to_mx_value(),
            style=style,
            parent="1",
            vertex="1",
        )
        self.add_mx_geo(node_element, node.rect)

    def get_node_by_sid(self, sid):
        try:
            return self.nodes[sid]
        except KeyError:
            return None

    @staticmethod
    def add_mx_geo(element, rect=None):
        if rect is None:
            ET.SubElement(element, MxConst.GEO, {"as": "geometry"}, relative="1")
        else:
            attributes = rect.to_dict_str()
            attributes["as"] = "geometry"
            ET.SubElement(element, MxConst.GEO, attributes)

    @staticmethod
    def add_mx_geo_with_points(element, curve):
        geo = ET.SubElement(element, MxConst.GEO, {"as": "geometry"}, relative="1")
        # TODO: needs to account for multiple bezier in path
        # array = ET.SubElement(geo, MxConst.ARRAY, {"as": "points"})
        # for i in MxConst.CURVE_INTERVALS:
        #     point = curve.cubic_bezier_coordinates(i)
        #     x, y = MxGraph.x_y_strs(point)
        #     ET.SubElement(array, MxConst.POINT, x=x, y=y)

    @staticmethod
    def x_y_strs(point):
        return str(int(point.real)), str(int(point.imag))

    def value(self):
        return MxConst.DECLARATION + ET.tostring(self.graph, encoding="unicode")



from . import db, Node, Edge, dbgraph
from ..mx.models.UserObject import UserObject

def node_is_user_object(node):
    value = node.value
    if (value is not None) and (type(value) == UserObject):
        return True
    return False

def get_node_trust_zones_from_graph(graph):
    zones = dict()
    nodes = set(graph.nodes.keys())    
    edges = set()
    for edge in graph.edges:
        edges.add(edge.to)
        edges.add(edge.fr)

    orphans = nodes.difference(edges)
    nodes = nodes.difference(orphans)
    
    # For now, orphans are assumed to be the smaller inner objects
    for node in nodes:
        node = graph.get_node_by_sid(node)

        for orphan in orphans: 
            zone = None
            orphan = graph.get_node_by_sid(orphan)
            
        
            outer_rect = node.rect
            inner_rect = orphan.rect
            
            # here, both the element node and zone node could be wrapped user objects
            if outer_rect.is_overlapping(inner_rect):
                
                # maybe using our shape library
                if node_is_user_object(node) and node_is_user_object(orphan):
                    assert(orphan.value.get_object_type() == 'trust zone')
                    zone = orphan.value.get_trust_zone()
                    
                # using the built-in shape library
                else:
                    zone = UserObject.get_trust_zone_from_node_label(orphan.value)

            if zone is not None:
                zones[node] = zone
                print("found {} {} inside {} {}".format(zones[node], orphan.sid, node.label, node.sid))

    return zones

def load_graph_into_db(graph, zones):
    flows = graph.edges.copy()
    
    remove_flow = lambda flow: flows.remove(flow)

    # we only care about things that are connected, so we start by traversing all of the edges in the graph
    for flow in flows:
        SOURCE = 0
        DESTINATION = 1

        source = graph.nodes[flow.fr]
        destination = graph.nodes[flow.to]
        
        # Processes get special treatment in our scheme. They represent metadata about a given flow
        process = None
        
        pair = []

        for node in (source, destination): 
            print("inspecting {} {}".format(node.label, node.sid))
            element_type = None

            if node_is_user_object(node):
                element_type = node.value.get_object_type()
            else:
                element_type = UserObject.infer_type_from_node(node)

            if element_type == 'process':
                process = Node().create(
                    zone=zones[node],
                    label=node.label,
                    identifier=node.sid,
                    data='test',
                    type=element_type
                )

                if node == source:
                    print("fix up source {} {}".format(node.label, node.sid))

                    inbound_flow, inbound_node = [(flow, flow.fr) for flow in flows if flow.to == node.sid][0]
                    node = graph.nodes[inbound_node]
                    dangling_flow = inbound_flow            

                else:
                    print("fix up destination {} {}".format(node.label, node.sid))
                    
                    outbound_flow, outbound_node = [(flow, flow.to) for flow in flows if flow.fr == node.sid][0]
                    node = graph.nodes[outbound_node]
                    dangling_flow = outbound_flow

                flows.remove(flow)
                flows.remove(dangling_flow)

            pair.append(
                Node().create(
                    zone=zones[node],
                    label=node.label,
                    identifier=node.sid,
                    data='test',
                    type=element_type
                )
            )

        Edge.create(
            source=pair[SOURCE],
            destination=pair[DESTINATION],
            process=process,
            data='test'
        )
    
    return True




from ..models.Curve import Curve


class CurveFactory:
    def __init__(self, coords):
        super(CurveFactory, self).__init__()
        self.coords = coords

    def from_xml(self, xml):
        geometry = xml.find('mxGeometry')
        points = geometry.findall('mxPoint')

        start = 0
        end = 0

        for point in points:
            x = float(point.get('x'))
            y = float(point.get('y'))
            
            if point.get('as') == 'sourcePoint':
                start = complex(x, y)
            elif point.get('as') == 'targetPoint':
                end = complex(x, y)

        
        return Curve(start=start, end=end, cb=None)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="materialize-threats",
    version="1.0.4",
    author="Jacob Salassi",
    author_email="author@example.com",
    description="Analyze draw.io data flow diagrams for STRIDE threat classes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/secmerc/materialize-threats",
    packages=setuptools.find_packages(),
    install_requires=[
          'peewee',
          'gherkin-official',
          'pytest'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    entry_points = {
        'console_scripts': ['materialize-threats=materialize_threats.materialize:ThreatMaterializer.materialize'],
    }
)



from ..io.CurveFactory import CurveFactory
from ..models.Edge import Edge


class EdgeFactory:
    def __init__(self, coords):
        super(EdgeFactory, self).__init__()
        self.curve_factory = CurveFactory(coords)

    def from_xml(self, xml, value):
        return Edge(
            sid=xml.get('id'),
            gid=xml.get('parent'),
            value=value,
            fr=xml.get('source'),
            to=xml.get('target'),
            curve=self.curve_factory.from_xml(xml)
        )


from . import LinearRegression

linear_min_r2 = 0.9


class Curve:
    def __init__(self, start, end, cb):
        """Takes complex numbers for start, end, and list of 4 Bezier control points"""
        self.start = start
        self.end = end
        assert cb is None or len(cb) == 4  # TODO: emit value error instead
        self.cb = cb

    def __str__(self):
        control = "[" + (str(self.cb) if self.cb is not None else "None") + "]"
        return str(self.start) + ", " + control + ", " + str(self.end)

    @staticmethod
    def is_linear(points, threshold=linear_min_r2):
        """
        Returns a boolean indicating whether a list of complex points is linear.

        Takes a list of complex points and optional minimum R**2 threshold for linear regression.
        """
        r2 = LinearRegression.coefficients(points)[2]
        return r2 > threshold

    def cubic_bezier_coordinates(self, t):
        """
        Returns a complex number representing the point along the cubic bezier curve.

        Takes parametric parameter t where 0 <= t <= 1
        """
        x = Curve._cubic_bezier(self._cb("real"), t)
        y = Curve._cubic_bezier(self._cb("imag"), t)
        return complex(x, y)

    def _cb(self, prop):
        return [getattr(x, prop) for x in self.cb]

    @staticmethod
    def _cubic_bezier(p, t):
        """
        Returns a float representing the point along the cubic bezier curve in the given dimension.

        Takes ordered list of 4 control points [P0, P1, P2, P3] and parametric parameter t where 0 <= t <= 1

        implements explicit form of https://en.wikipedia.org/wiki/B%C3%A9zier_curve#Cubic_B%C3%A9zier_curves
        """
        assert 0 <= t <= 1  # TODO: emit value error instead
        return (
            (((1.0 - t) ** 3) * p[0])
            + (3.0 * t * ((1.0 - t) ** 2) * p[1])
            + (3.0 * (t ** 2) * (1.0 - t) * p[2])
            + ((t ** 3) * p[3])
        )
